body{margin:0;padding:0;font-family:-apple-system,BlinkMacSystemFont,Segoe UI,Roboto,Oxygen,Ubuntu,Cantarell,Fira Sans,Droid Sans,Helvetica Neue,sans-serif;-webkit-font-smoothing:antialiased;-moz-osx-font-smoothing:grayscale}code{font-family:source-code-pro,Menlo,Monaco,Consolas,Courier New,monospace}.App,.bottom{width:100%;height:-webkit-fit-content;height:-moz-fit-content;height:fit-content}.bottom{display:-webkit-flex;display:flex;float:left}.overview{width:800px;height:600px}.clusters,.overview{border:solid;border-width:thin}.clusters{width:50%;height:908px;overflow-y:scroll}.eachCluster{width:98%;height:-webkit-fit-content;height:-moz-fit-content;height:fit-content;border:solid;border-width:thin;margin:3px;padding:5px;font-size:small}.rv-xy-plot{display:-webkit-flex;display:flex}.citing{width:800px;height:-webkit-fit-content;height:-moz-fit-content;height:fit-content;border:solid;border-width:thin;display:block}.abstract{width:650px;margin:auto}.map{width:790px;height:400px;display:block}.citationlist{width:100%;height:500px;overflow-y:scroll}.citationlistExtended{width:100%;height:908px;overflow-y:scroll}.hfill{width:-webkit-fit-content;width:-moz-fit-content;width:fit-content;margin:auto 10px auto auto}.mapboxgl-map{width:790px;height:400px;border:solid;margin:1px;border-color:#a9a9a9;border-width:medium;float:left}.year-control{border:solid;border-width:thin;background:#fff;padding:8px;border-radius:4px;margin:8px}
/*# sourceMappingURL=main.13b54120.chunk.css.map */